import collections

space = b" "[0]
offsets = (8, 28, 36, 44, 49, 55)
bcpColumn = collections.namedtuple(
    "bcpColumn", ["type", "bytes_indicator", "bytes_data", "column_name", "collation"]
)
bcpVersion = b"14.0"
collation_utf8 = "LATIN1_GENERAL_100_CI_AS_SC_UTF8"
collation_default = '""'


def load(format_bytes: bytes) -> list[bcpColumn]:
    "Parse format and run some checks ensuring correctness."
    version, nr_lines, *lines = format_bytes.splitlines()
    float(version)
    assert len(lines) == int(nr_lines.decode())
    spacers = set.intersection(
        *[
            set(
                i for i in range(len(line)) if line[i] != space and line[i - 1] == space
            )
            for i_line, line in enumerate(lines)
        ]
    )
    assert len(spacers) == 7
    assert spacers - {max(spacers)} == set(offsets)
    spacers = sorted(spacers)
    columns = []
    for i_line, line in enumerate(lines):
        assert line[0] != space
        assert line[-1] != space
        parts = [line[start:end] for start, end in zip([0] + spacers, spacers + [None])]
        assert int(parts[0].decode().strip()) == i_line + 1
        assert int(parts[5].decode().strip()) == i_line + 1
        col_args = [
            i.strip().decode(errors="ignore")
            for i in [parts[1], parts[2], parts[3], parts[6], parts[7]]
        ]
        col_args[1] = int(col_args[1])
        col_args[2] = int(col_args[2])
        columns.append(bcpColumn(*col_args))
    return columns


def dump(bcp_columns: list[bcpColumn]) -> bytes:
    "Convert to content of a bcp format file"
    # name_padding = max(len(i.column_name) for i in bcp_columns) + 16
    name_padding = len(bcp_columns) + 16
    name_padding = int(name_padding // 4 * 4)
    just = [
        offsets[0],
        offsets[1] - offsets[0],
        offsets[2] - offsets[1],
        offsets[3] - offsets[2],
        offsets[4] - offsets[3],
        offsets[5] - offsets[4],
        name_padding,
    ]
    return b"\n".join(
        [bcpVersion, str(len(bcp_columns)).encode()]
        + [
            b"".join(
                [
                    str(i_col).encode().ljust(just[0]),
                    col.type.encode().ljust(just[1]),
                    str(col.bytes_indicator).encode().ljust(just[2]),
                    str(col.bytes_data).encode().ljust(just[3]),
                    b'""'.ljust(just[4]),
                    str(i_col).encode().ljust(just[5]),
                    f"column_{i_col}".encode().ljust(just[6]),
                    col.collation.encode(),
                ]
            )
            for i_col, col in enumerate(bcp_columns, 1)
        ]
        + [b""]
    )


def native_to_implemented_types(bcp_columns: list[bcpColumn]) -> list[bcpColumn]:
    "Convert SQL types to ones that are closer to arrow types"
    converted_bcp_columns = []
    for col in bcp_columns:
        new_type = col.type
        new_bytes_indicator = 1  # always request null indicator
        new_bytes_data = col.bytes_data
        new_collation = collation_default
        match col.type:
            case "SQLBINARY" | "SQLUDT":
                new_type = "SQLBINARY"
                new_bytes_data = 0  # or varbinary(max)
                new_bytes_indicator = 8
            case "SQLCHAR" | "SQLNCHAR" | "SQLVARIANT":
                new_type = "SQLCHAR"
                new_bytes_data = 0  # request as varchar(max)
                new_bytes_indicator = 8
                new_collation = collation_utf8
            case "SQLDATETIM4" | "SQLDATETIME":
                new_type = "SQLDATETIME2"
                new_bytes_data = 8
            case "SQLMONEY" | "SQLMONEY4" | "SQLNUMERIC":
                new_type = "SQLDECIMAL"
                new_bytes_data = 19
        converted_bcp_columns.append(
            bcpColumn(
                type=new_type,
                bytes_indicator=new_bytes_indicator,
                bytes_data=new_bytes_data,
                column_name=col.column_name,
                collation=new_collation,
            )
        )
    return converted_bcp_columns



from setuptools.command.build_ext import build_ext
from setuptools import Extension, setup
import sys
import subprocess
from pathlib import Path
import platform
import sysconfig
import shutil

module_name = "arrow_bcp"

class ZigBuilder(build_ext):
    def build_extension(self, ext):
        source, = ext.sources
        build_path = Path(self.build_lib, module_name)
        build_path.mkdir(parents=True, exist_ok=True)

        def escape(path):
            return path.replace("\\", "\\\\")
        
        lib_paths = [str(Path(sysconfig.get_config_var('installed_base'), "Libs").absolute())]
        with Path("src-zig", "include_dirs.zig").open("w", encoding="utf-8") as f:
            f.write(
                f"pub const include: [{len(self.include_dirs)}][]const u8 = .{{\n"
                + "".join(f'    "{p}",\n' for p in map(escape, self.include_dirs))
                + "};\n"
                + f"pub const lib: [{len(lib_paths)}][]const u8 = .{{\n"
                + "".join(f'    "{p}",\n' for p in map(escape, lib_paths))
                + "};\n"
            )

        windows = platform.system() == "Windows"

        subprocess.call([
            sys.executable,
            "-m",
            "ziglang",
            "build",
            *(["-Dtarget=x86_64-windows"] if windows else []),
            "--release=safe",
        ], cwd=source)

        binary, = (p for p in Path("src-zig", "zig-out").glob(f"**/*{'.dll' if windows else ''}") if p.is_file())
        shutil.copyfile(binary, build_path / self.get_ext_filename(ext.name))

setup(
    ext_modules=[Extension("zig_ext", ["src-zig"])],
    cmdclass={"build_ext": ZigBuilder},
)



import typing
import tempfile
from pathlib import Path
import subprocess
import pyarrow as pa

from . import bcp_format, bcp_data

bcp_exe = "bcp"


def set_bcp_executable(location: str):
    global bcp_exe
    bcp_exe = location


class BcpError(subprocess.SubprocessError):
    "Raised when a bcp subprocess fails"
    pass


class BcpReader:
    "Used by fetch_iter, don't invoke directly"

    def __init__(
        self,
        path_data: Path,
        tempdir_read: tempfile.TemporaryDirectory,
        bcp_columns: list[bcp_format.bcpColumn],
        batch_size: int,
    ):
        self.path_data = path_data
        self.tempdir_read = tempdir_read
        self.bcp_columns = bcp_columns
        self.has_entered = False
        self.batch_size = batch_size

    def __iter__(self) -> typing.Generator[pa.RecordBatch, None, None]:
        if not self.has_entered:
            raise Exception("Must use from context manager")
        return bcp_data.load(
            path_data=self.path_data,
            bcp_columns=self.bcp_columns,
            max_rows=self.batch_size,
        )

    def __enter__(self):
        self.tempdir_read.__enter__()
        self.has_entered = True
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.tempdir_read.__exit__(exc_type, exc_value, traceback)


def _run_proc(args, operation):
    proc = subprocess.run(args, capture_output=True)
    try:
        proc.check_returncode()
    except subprocess.SubprocessError as e:
        raise BcpError(
            f"BCP process to {operation} terminated with error",
            proc.stdout,
            proc.stderr,
        ) from e
    else:
        if proc.stderr:
            raise BcpError(
                f"BCP process to {operation} received unexpected stderr",
                proc.stdout,
                proc.stderr,
            )
        return proc.stdout


class ConnectionInfo:
    """
    bcp_args gets unpacked directly into the bcp subprocess call. See Microsoft's
    official bcp.exe documentation for authentication options.

    The path to the bcp executable is defined on the package level and can be
    modified elsewhere.

    Arguments regarding table name/format like "in", "out", "format", "-f" "-n"
    are managed by arrow_bcp.

    Consider passing "-h" "TABLOCK" to potentially speed up inserts.
    """

    def __init__(self, bcp_args: typing.Iterable):
        self.bcp_args = bcp_args

    def download_format_file(self, path_format: Path, table: str) -> str:
        """
        Execute a bcp process to download the native format file for the specified table.
        Return stdout on success or stdout + stderr as part of an Exception on failure.
        """
        return _run_proc(
            [
                bcp_exe,
                table,
                "format",
                "nul",
                "-f",
                str(path_format),
                *self.bcp_args,
                "-n",
            ],
            "download a format file",
        )

    def download_data_file(self, path_format: Path, path_data: Path, table: str) -> str:
        """
        Execute a bcp process to download the data file based on the specified format file.
        Return stdout on success or stdout + stderr as part of an Exception on failure.
        """
        return _run_proc(
            [
                bcp_exe,
                table,
                "out",
                str(path_data),
                "-f",
                str(path_format),
                *self.bcp_args,
            ],
            "download data",
        )

    def insert_file(self, path_format: Path, path_data: Path, table: str) -> str:
        """
        Execute a bcp process to insert the data from the data/format files into the specified table.
        Return stdout on success or stdout + stderr as part of an Exception on failure.
        """
        return _run_proc(
            [
                bcp_exe,
                table,
                "in",
                str(path_data),
                "-f",
                str(path_format),
                *self.bcp_args,
            ],
            "insert data",
        )

    def insert_arrow(
        self, table: str, arrow_data: pa.Table | typing.Iterable[pa.RecordBatch]
    ):
        "Bulk insert arrow table or record batches into table"
        with tempfile.TemporaryDirectory(prefix="arrow_bcp_write_") as tmpdir_dl_name:
            path_fmt = Path(tmpdir_dl_name) / "bcp_download.fmt"
            path_dat = Path(tmpdir_dl_name) / "bcp_download.dat"
            if isinstance(arrow_data, pa.Table):
                arrow_data = arrow_data.to_batches()
            if bcp_data.dump(
                batches=arrow_data, path_format=path_fmt, path_data=path_dat
            ):
                self.insert_file(path_format=path_fmt, path_data=path_dat, table=table)

    def download_arrow_batches(self, table: str, batch_size: int = 500) -> BcpReader:
        """
        Download data for a table and return iterable of arrow batches to allow lazy loading.
        The returned reader object must be used from a context manager. Upon exiting the
        reader object, the temporary BCP data file will be cleaned up.
        """
        with tempfile.TemporaryDirectory(prefix="arrow_bcp_read_") as tmpdir_dl_name:
            path_fmt = Path(tmpdir_dl_name) / "bcp_download.fmt"
            self.download_format_file(path_fmt, table)
            with path_fmt.open("rb") as f:
                bcp_columns = bcp_format.load(f.read())

            bcp_columns = bcp_format.native_to_implemented_types(bcp_columns)

            with path_fmt.open("wb") as f:
                f.write(bcp_format.dump(bcp_columns))

            path_dat = Path(tmpdir_dl_name) / "bcp_download.dat"
            self.download_data_file(
                path_format=path_fmt, path_data=path_dat, table=table
            )

            tmpdir_read = tempfile.TemporaryDirectory(prefix="arrow_bcp_read_")
            path_dat.rename(Path(tmpdir_read.name) / "bcp_download.dat")
            return BcpReader(
                path_data=Path(tmpdir_read.name) / "bcp_download.dat",
                tempdir_read=tmpdir_read,
                bcp_columns=bcp_columns,
                batch_size=batch_size,
            )

    def download_arrow_table(self, table: str, batch_size: int = 500) -> pa.Table:
        """
        Wrapper around download_arrow_batches returning a singular table. The table's
        entire data is loaded at once.
        """
        with self.download_arrow_batches(table) as reader:
            batches = [*reader]
        target_schema = batches[-1].schema
        for i_batch, batch in enumerate(batches[:-1]):
            if batch.schema != target_schema:
                batches[i_batch] = batch.cast(target_schema)
            else:
                # once we figured out the schema it will not change anymore
                break
        return pa.Table.from_batches(batches)



import pyarrow as pa
from pathlib import Path
import typing
import itertools
from . import zig_ext, bcp_format


def array_capsule_generator(batch_iter: typing.Iterable[pa.RecordBatch], i_col: int):
    dtype_prev = None
    for batch in batch_iter:
        col = batch.columns[i_col]
        assert dtype_prev is None or col.type == dtype_prev
        dtype_prev = col.type
        _, capsule_array = col.__arrow_c_array__()
        yield capsule_array


def dump(
    batches: typing.Iterable[pa.RecordBatch], path_format: Path, path_data: Path
) -> bool:
    """
    Write arrow data to bcp files. It is assumed that each batch has the same schema.
    Return True if any data was written and False if no batches got submitted.
    """
    batch_iter = iter(batches)
    try:
        first_batch = next(batch_iter)
    except StopIteration:
        return False
    names = first_batch.column_names
    schema_capsules = [arr.__arrow_c_array__()[0] for arr in first_batch.columns]
    batch_iterators = itertools.tee(
        itertools.chain([first_batch], batch_iter), len(names)
    )
    array_capsules = [
        array_capsule_generator(it, i_col) for i_col, it in enumerate(batch_iterators)
    ]

    path_data = path_data.absolute()
    assert path_data.parent.exists()
    sql_info = zig_ext.write_arrow(schema_capsules, array_capsules, str(path_data))

    bcp_columns = [
        bcp_format.bcpColumn(
            type=tp,
            bytes_indicator=bi,
            bytes_data=bd,
            column_name=name,
            collation=(
                bcp_format.collation_utf8
                if tp == "SQLCHAR"
                else bcp_format.collation_default
            ),
        )
        for (tp, bi, bd), name in zip(sql_info, names, strict=True)
    ]
    with path_format.open("wb") as f:
        f.write(bcp_format.dump(bcp_columns=bcp_columns))
    return True


def load(
    bcp_columns: list[bcp_format.bcpColumn], path_data: Path, max_rows: int
) -> typing.Generator[pa.RecordBatch, None, None]:
    "Load a bcp data file into arrow batches according to the specified format."
    names = [col.column_name for col in bcp_columns]

    reader_state = zig_ext.init_reader(
        [(col.type, col.bytes_indicator, col.bytes_data) for col in bcp_columns],
        str(path_data.absolute()),
    )

    go_again = True
    while go_again:
        capsules = zig_ext.read_batch(reader_state, max_rows)
        arrays = [
            pa.Array._import_from_c_capsule(schema, array) for schema, array in capsules
        ]
        batch = pa.record_batch(arrays, names=names)
        go_again = len(batch) == max_rows
        yield batch
